#! /usr/bin/env python

import sys
import mistral_lib
import types


#from Numberjack import *


try:
    from Numberjack import VarArray, Matrix, ConstraintNotSupportedError, val_heuristics, var_heuristics
except:
    pass

UNSAT     =  0;
SAT       =  1;
UNKNOWN   =  2;
LIMITOUT  =  3;
LUBY      =  0; 
GEOMETRIC =  1;

#val_heuristics = ['Lex', 'AntiLex', 'Random', 'RandomMinMax', 'DomainSplit', 'No']
#var_heuristics = ['MinDomain', 'Lex', 'MaxDegree', 'MinDomainMinVal', 'Random', 'MinDomainMaxDegree', 'DomainOverDegree', 'DomainOverWDegree', 'DomainOverWLDegree', 'Neighbour', 'Impact', 'ImpactOverDegree', 'ImpactOverWDegree', 'ImpactOverWLDegree', 'Scheduling']



#============   Util    =============#
def toVarArray(l):
    nItems = len(l)
    Y = mistral_lib.new_VarArray((nItems+1))
    for i in range(nItems):
        mistral_lib.VarArray_setitem(Y,i,l[i].var_pointer)
    return Y

def toIntArray(l, offset=1):
    nItems = len(l)
    Y = mistral_lib.new_IntArray(nItems+offset)
    for i in range(nItems):
        mistral_lib.IntArray_setitem(Y,i,l[i])
    for i in range(offset):
        mistral_lib.IntArray_setitem(Y,nItems+i,0)
    return Y



#============   Model   =============#
class Model(object):
    def __init__(self):        
        self.mistral_model = mistral_lib.CSP()
        self.solver = None
    def add(self, X):
        if getattr(X, '__iter__', False):
            for x in X:
                self.add( x )
        else:
            self.mistral_model.add( X.var_pointer )
        #self.mistral_model.add( X.var_pointer )
    def __iadd__(self, X):
        self.mistral_model.add( X.var_pointer )
    def __str__(self):
        if self.solver == None:
            self.mistral_model.printPython()
        else:
            self.solver.printPython()
        return ' '

#============  Solver   =============#
#class Solver(NBJ_STD_Solver):
class Solver(object):
    def __init__(self, model, X=None):
        #Numberjack.NBJ_STD_Solver.__init__(self, model, "Mistral", X)

        self.library = 'Mistral'
        var_array = None
        tmp_array = []
        if type(model) == Model:
            self.wrapper_model = model
            if X != None:
                var_array = toVarArray( X )
        else:
            self.wrapper_model = Model() #mistral.CSP()            
            self.load_in_global_model(model)
            if X != None:
                for x in X:
                    if hasattr(x, 'var'):
                        tmp_array.append(x.var)
                var_array = toVarArray( tmp_array )
                #var_array = toVarArray( [x.var for x in X] )
                #print [x.ident for x in X]

        if var_array == None:
            self.solver = mistral_lib.Solver( self.wrapper_model.mistral_model )
        else:
            self.solver = mistral_lib.Solver( self.wrapper_model.mistral_model, var_array, len(tmp_array) )
            mistral_lib.delete_VarArray( var_array )

    def delete(self):
        mistral_lib._mistral_lib.delete_CSP( self.wrapper_model.mistral_model )
        mistral_lib._mistral_lib.delete_Solver( self.solver )

    def load_in_global_model(self, gm):
        self.model = gm
        self.model.close()
        for expr in gm.get_exprs():
            self.wrapper_model.add(self.load_expr(expr))

    def load_expr(self, expr):
        if type(expr) is str:
            return self.model.string_map[expr]
        if type(expr) is type(1):
            # It is a constant, handle appropriatly
            return expr # Gecode_Constant(expr)
        if not expr.has_children():
            # It is a leaf
            if expr.is_var(): # Just to be sure
                alreadyBuild = False
                if expr.is_build():
                    if expr.get_solver() == self: # Checks if I have already assigned
                        alreadyBuild = True

                if alreadyBuild:
                    return expr.var
                else:
                    # It is probably a variable
                    (lb,ub,domain) = expr.get_domain()
                    
                    var = None
                    if domain is None:
                        var = Variable(lb, ub)
                    elif ub-lb+1 == len(domain):
                        var = Variable(lb, ub)
                    else:
                        var = Variable(domain)

                    expr.var = var # Record that the var has been created
                    expr.solver = self
                    return var
            else:
                raise Error("Problem, no such type exists in converting models")
        else:
            #print ("Mistral_%s" % expr.get_operator())
            factory = getattr(sys.modules[__name__],
                              "Mistral_%s" % expr.get_operator(), None)


            if factory is not None:

                arguments = [self.load_expr(child) for child in expr.get_children()]

                if expr.has_parameters(): # != None: # assumes an array of integers
                    expr.var = factory(arguments,expr.parameters)
                else:
                    expr.var = factory(arguments)
                expr.solver = self
                return expr.var
            
            else:
                return self.decompose_expression(expr)
        
    def decompose_expression(self, expr):
        if hasattr(expr, "decompose"):
            expr_list = expr.decompose()
            for exp in expr_list[1:]:
                self.wrapper_model.add(self.load_expr(exp))
            var = self.load_expr(expr_list[0])
            expr.solver = self
            expr.var = var
            return var
        else:
            print "ERROR: Solver does not support \"%s\" and no decomposition is available" % (expr.get_operator())
            raise ConstraintNotSupportedError(expr.get_operator(), "Mistral")

    # solving methods
    def solve(self):
        print 'c  solve using Mistral'
        try:
            if self.solver.solve() == SAT:
                return True
            elif self.solver.SOLUTIONS > 0:
                return True
            return False        
        except (KeyboardInterrupt, SystemExit):
            print "Caught signal"
            return
    def ldSolve(self, ideal=None, scope=None, step=1, limit=sys.maxint):
        print 'c  solve using Mistral'
        try:
            if ideal == None:
                if self.solver.ldSolve() == SAT:
                    return True
            else:
                int_array = toIntArray(ideal)
                if scope == None:
                    if self.solver.ldSolve(int_array) == SAT:
                        return True
                else:
                    var_array = toVarArray( [x.var for x in scope] )
                    if self.solver.ldSolve(len(scope), var_array, int_array, step, limit) == SAT:
                        return True
            return False
        except (KeyboardInterrupt, SystemExit):
            return 
    def restart(self):
        self.solver.restart() 
    def solveAndRestart(self, policy=GEOMETRIC, base=64, factor=1.3):
        print 'c  solve using Mistral'
        try:
            if self.solver.solve_and_restart(policy,base,factor) == SAT:
                return True
            elif self.solver.SOLUTIONS > 0:
                return True
            return False
        except (KeyboardInterrupt, SystemExit):
            return
    def startNewSearch(self):
        print 'c  solve using Mistral'
        self.solver.startNewSearch()
    def getNextSolution(self):
        try:
            return self.solver.getNextSolution()
        except (KeyboardInterrupt, SystemExit):
            return 
    def sacPreprocess(self,type):
        self.solver.sacPreprocess(type)
        return (self.solver.status != UNSAT)

    # parameter tuning methods
    def setBranching(self, vars, branching):
        tmp_array = []
        for x in vars:
            if hasattr(x, 'var'):
                tmp_array.append(x.var)
        var_array = toVarArray( tmp_array )
        #var_array = toVarArray( [x.var for x in vars] )
        #var_array = toVarArray(vars)
        self.solver.setBranching(var_array, len(tmp_array), branching)

    def setHeuristic(self, arg1, arg2='No', arg3=0):
        var_name = arg1
        val_name = arg2
        randomization = arg3
        if type(val_name) is int:
            randomization = val_name
            val_name = 'No'
        if var_name not in var_heuristics:
            print 'c Warning: "'+var_name+'" unknown, use MinDomain instead' 
            print 'c legal variable orderings: ', var_heuristics
        if val_name not in val_heuristics:
            print 'c Warning: "'+val_name+'" unknown, use Lex instead' 
            print 'c legal value orderings: ', val_heuristics
        self.solver.setHeuristic(str(var_name), str(val_name), randomization)
    def setFailureLimit(self,cutoff):  
        self.solver.setFailureLimit(cutoff)
    def setTimeLimit(self,cutoff):
        self.solver.setTimeLimit(cutoff)
    def setVerbosity(self, degree):
        self.solver.setVerbosity( degree )
    def setRandomized(self, degree):
        self.solver.setRandomized(degree)
    def setRandomSeed(self, seed):
        self.solver.setRandomSeed(seed);

    # statistics methods
    def is_sat(self):
        return self.solver.status == SAT
    def is_unsat(self):
        return self.solver.status == UNSAT
    def printStatistics(self):
        print ''
        self.solver.printStatistics()
        print ''
    def getBacktracks(self):
        return self.solver.getBacktracks();
    def getNodes(self):
        return self.solver.getNodes();
    def getFailures(self):
        return self.solver.getFailures();
    def getChecks(self):
        return self.solver.getChecks();
    def getPropags(self):
        return self.solver.getPropags();
    def getTime(self):
        return self.solver.getTime();

    def __str__(self):
        self.solver.printPython()
        return ' '

#============ Variables =============#

class Variable:
    # constructor(s)
    def __init__(self, lb=None, ub=None):
        if lb == None:
            self.var_pointer = None
        elif ub == None:
            if type(lb) is int:
                self.var_pointer = mistral_lib.CSP._Variable(lb, lb)
            else:
                domain = sorted(lb)
                lb = domain[0]
                ub = domain[len(domain)-1]
                val_array = toIntArray(domain, 0)
                self.var_pointer = mistral_lib.CSP._Variable(val_array, len(domain), lb, ub)
        else:
            self.var_pointer = mistral_lib.CSP._Variable(lb, ub)

    def get_value(self):
        return self.var_pointer.getValue()

    def get_min(self):
        return self.var_pointer.getMin()

    def get_max(self):
        return self.var_pointer.getMax()

    def __str__(self):
        return str(self.get_value())

    # arythmetic relations
    def __lt__(self, other):
        return Mistral_lt([self,other])

    def __le__(self, other):
        return Mistral_le([self,other])

    def __gt__(self, other):
        return Mistral_gt([self,other])

    def __ge__(self, other):
        return Mistral_ge([self,other])

    def __eq__(self, other):
        return Mistral_eq([self,other])

    def __ne__(self, other):
        return Mistral_ne([self,other])


    # arythmetic operations
    def __add__(self, other):
        return Mistral_add([self,other])

    def __sub__(self, other):
        return Mistral_sub([self,other])

    def __mul__(self, other):
        return Mistral_mul([self,other])

    def __div__(self, other):
        return Mistral_div([self,other])

    def __mod__(self, other):
        return Mistral_mod([self,other])

    def __neg__(self):
        return Mistral_neg([self])


    # logic predicates
    def __or__(self, other):
        return Mistral_or([self,other])

    def __and__(self, other):
        return Mistral_and([self,other])

    def __not__(self):
        return Mistral_not([self])


def Mistral_lt(vars):
    X = Variable()
    if type(vars[1]) == int:
        X.var_pointer = mistral_lib.CSP._Precedence(vars[0].var_pointer, vars[1]-1)
    elif type(vars[0]) == int:
        X.var_pointer = mistral_lib.CSP._Precedence(vars[0]+1, vars[1].var_pointer)
    else:
        X.var_pointer = mistral_lib.CSP._Precedence(vars[0].var_pointer, 1, vars[1].var_pointer)
    return X

def Mistral_le(vars):
    X = Variable()
    if type(vars[1]) == int:
        X.var_pointer = mistral_lib.CSP._Precedence(vars[0].var_pointer, vars[1])
    elif type(vars[0]) == int:
        X.var_pointer = mistral_lib.CSP._Precedence(vars[0], vars[1].var_pointer)
    else:
        X.var_pointer = mistral_lib.CSP._Precedence(vars[0].var_pointer, 0, vars[1].var_pointer)
    return X

def Mistral_gt(vars):
    X = Variable()
    if type(vars[1]) == int:
        X.var_pointer = mistral_lib.CSP._Precedence(vars[1]+1, vars[0].var_pointer)
    elif type(vars[0]) == int:
        X.var_pointer = mistral_lib.CSP._Precedence(vars[1].var_pointer, vars[0]-1)
    else:
        X.var_pointer = mistral_lib.CSP._Precedence(vars[1].var_pointer, 1, vars[0].var_pointer)
    return X

def Mistral_ge(vars):
    X = Variable()
    if type(vars[1]) == int:
        X.var_pointer = mistral_lib.CSP._Precedence(vars[1], vars[0].var_pointer)
    elif type(vars[0]) == int:
        X.var_pointer = mistral_lib.CSP._Precedence(vars[1].var_pointer, vars[0])
    else:
        X.var_pointer = mistral_lib.CSP._Precedence(vars[1].var_pointer, 0, vars[0].var_pointer)
    return X

def Mistral_eq(vars):
    X = Variable()
    if type(vars[1]) == int:
        X.var_pointer = mistral_lib.CSP._Equal(vars[0].var_pointer, vars[1], 1)
    elif type(vars[0]) == int:
        X.var_pointer = mistral_lib.CSP._Equal(vars[1].var_pointer, vars[0], 1)
    else:
        X.var_pointer = mistral_lib.CSP._Equal(vars[0].var_pointer, vars[1].var_pointer, 1)
    return X

def Mistral_ne(vars):
    X = Variable()
    if type(vars[1]) == int:
        X.var_pointer = mistral_lib.CSP._Equal(vars[0].var_pointer, vars[1], 0)
    elif type(vars[0]) == int:
        X.var_pointer = mistral_lib.CSP._Precedence(vars[1].var_pointer, vars[0], 0)
    else:
        X.var_pointer = mistral_lib.CSP._Equal(vars[0].var_pointer, vars[1].var_pointer, 0)
    return X

# arythmetic operations
def Mistral_add(vars):
    X = Variable()
    if type(vars[1]) == int:
        if vars[1] != 0:
            X.var_pointer = mistral_lib.CSP._Sum(vars[0].var_pointer, vars[1])
        else:
            X.var_pointer = vars[0].var_pointer
    else:
        X.var_pointer = mistral_lib.CSP._Sum(vars[0].var_pointer, vars[1].var_pointer, 1)
    return X

def Mistral_sub(vars):
    X = Variable()
    if type(vars[1]) == int:
        if vars[1] != 0:
            X.var_pointer = mistral_lib.CSP._Sum(vars[0].var_pointer, -vars[1])
        else:
            X.var_pointer = vars[0].var_pointer
    else:
        X.var_pointer = mistral_lib.CSP._Sum(vars[0].var_pointer, vars[1].var_pointer, -1)
    return X

def Mistral_mul(vars):
    X = Variable()
    if type(vars[1]) == int:
        if vars[1] != 1:
            X.var_pointer = mistral_lib.CSP._Mul(vars[0].var_pointer, vars[1])
        else:
            X.var_pointer = vars[0].var_pointer
    else:
        X.var_pointer = mistral_lib.CSP._Mul(vars[0].var_pointer, vars[1].var_pointer)
    return X

def Mistral_div(vars):
    X = Variable()
    if type(vars[1]) == int:
        if vars[1] != 1:
            X.var_pointer = mistral_lib.CSP._Div(vars[0].var_pointer, vars[1])
        else:
            X.var_pointer = vars[0].var_pointer
    else:
        X.var_pointer = mistral_lib.CSP._Div(vars[0].var_pointer, vars[1].var_pointer)
    return X

def Mistral_mod(vars):
    X = Variable()
    if type(vars[1]) == int:
        X.var_pointer = mistral_lib.CSP._Mod(vars[0].var_pointer, vars[1])
    else:
        X.var_pointer = mistral_lib.CSP._Mod(vars[0].var_pointer, vars[1].var_pointer)
    return X

def Mistral_neg(vars):
    X = Variable()
    X.var_pointer = mistral_lib.CSP._Neg(vars[0].var_pointer)
    return X

def Mistral_and(vars):
    X = Variable()
    X.var_pointer = mistral_lib.CSP._And(vars[0].var_pointer, vars[1].var_pointer)
    return X

def Mistral_or(vars):
    X = Variable()
    X.var_pointer = mistral_lib.CSP._Or(vars[0].var_pointer, vars[1].var_pointer)
    return X

def Mistral_not(vars):
    X = Variable()
    X.var_pointer = mistral_lib.CSP._Not(vars[0].var_pointer)
    return X


class Mistral_Table(Variable):
    def __init__(self, vars, tuples=[], type='support'):
        var_array = toVarArray(vars)
        self.var_pointer = mistral_lib.CSP._Table( var_array, len(vars), (type == 'support') )    
        for tuple in tuples:
            int_tuple = toIntArray(tuple, 0)
            mistral_lib.CSP._addTuple( self.var_pointer, int_tuple )



#============ Constraints =============#

def Mistral_AllDiff(vars, propag=None):
    var_array = toVarArray(vars)
    X = Variable()
    if propag is None: propag = [3]
    X.var_pointer = mistral_lib.CSP._AllDifferent( var_array, len(vars), propag[0], 0 )
    return X

def Mistral_Gcc(vars, params): #vals, lb, ub, propag=None):
    vals = params[0]
    lb = params[1]
    ub = params[2]
    var_array = toVarArray(vars)
    minval = vals[0]
    maxval = vals[len(vals)-1]
    inf = [0]*(maxval-minval+1)
    sup = [0]*(maxval-minval+1)
    for i in range(len(vals)):
        inf[vals[i]+minval] = lb[i]
        sup[vals[i]+minval] = ub[i]
    
    clb = toIntArray(inf)
    cub = toIntArray(sup)

    X = Variable()
    X.var_pointer = mistral_lib.CSP._Gcc( var_array, len(vars), minval, maxval, clb, cub )
    return X

def Mistral_Sum(vars, weights):
    if len(vars) > 0:
        var_array = toVarArray(vars)
        X = Variable()
        m_weights = weights[0]
        m_weights.append(weights[1])

        #print m_weights
        #exit(0)

        weight_array = toIntArray(m_weights)
        X.var_pointer = mistral_lib.CSP._Sum( var_array, len(vars), weight_array )
        mistral_lib.delete_IntArray(weight_array)
        return X
    else:
        return weights[1]
    

def Mistral_Min(vars):
    var_array = toVarArray(vars)
    X = Variable()
    X.var_pointer = mistral_lib.CSP._Min( var_array, len(vars) )
    return X

def Mistral_Max(vars):
    var_array = toVarArray(vars)
    X = Variable()
    X.var_pointer = mistral_lib.CSP._Max( var_array, len(vars) )
    return X

def Mistral_LexOrder(vars_1, vars_2, type):
    vars = None
    if vars_2 is not None:
        vars = list(vars_1)
        vars.append(vars_2)
    else:
        vars = vars_1
    var_array = toVarArray(vars)
    X = Variable()
    X.var_pointer = mistral_lib.CSP._LexOrder( var_array, len(vars), type )
    return X

def Mistral_LessLex(vars_1, vars_2=None):
    return Mistral_LexOrder(vars_1, vars_2, 1)

def Mistral_LeqLex(vars_1, vars_2=None):
    return Mistral_LexOrder(vars_1, vars_2, 0)

def Mistral_Element(vars, index=None):
    _vars = None
    if index is not None:
        _vars = list(vars)
        _vars.append(index)
    else:
        _vars = vars
    var_array = toVarArray(_vars)
    X = Variable()
    X.var_pointer = mistral_lib.CSP._Element( var_array, len(_vars), 0 )
    return X

def Mistral_Minimise(var):
    X = Variable()
    if getattr(var, '__iter__', False):
        X.var_pointer = mistral_lib.CSP._Minimise( var[0].var_pointer )
    else:
        X.var_pointer = mistral_lib.CSP._Minimise( var.var_pointer )
    return X

def Mistral_Maximise(var):
    X = Variable()
    if getattr(var, '__iter__', False):
        X.var_pointer = mistral_lib.CSP._Maximise( var[0].var_pointer )
    else:
        X.var_pointer = mistral_lib.CSP._Maximise( var.var_pointer )
    return X




class Task(Variable):
    def __init__(self, arg1=None, arg2=None, arg3=None):
        lb = 0
        ub = 1
        self.duration = 1
        if arg1 != None:
            if arg2 != None:
                if arg3 != None:
                    lb = arg1
                    ub = arg2-arg3
                    self.duration = arg3
                else: # only 2 args, read as <makespan,duration>
                    ub = arg1-arg2
                    self.duration = arg2
            else: # only 1 arg, read as the makespan
                ub = arg2-1
        Variable.__init__(self, lb, ub)
        
    def __lt__(self, pred):
        print 'add precedence with', pred
        if type(pred) == int:
            return mistral_lib.CSP._Precedence(self, pred - self.duration)
        else:
            return Precedence([self, pred], self.duration)
    
    def __gt__(self, pred):
        return pred.__lt__(self)

    def requires(self, resource):
        resource.add(self)

    def reset(self, makespan):
        print 'reset', makespan
        self.ub = makespan-self.duration

'''
class UnaryResource(list):
    def __init__(self, arg):
        list.__init__(self, [NoOverlap(arg[i], arg[j]) for i in range(1,len(arg)) for j in range(i)])
        self.tasks = [task for task in arg]

    def add(self, new_task):
        for task in self.tasks:
            self.append( NoOverlap(new_task, task) )
        self.tasks.append(new_task)
'''

def Mistral_Precedence(tasks, duration):
    X = Variable()
    if type(tasks[1]) is int:
        X.var_pointer = mistral_lib.CSP._Precedence(tasks[0].var_pointer, tasks[1] - duration[0])
    else:
        X.var_pointer = mistral_lib.CSP._Precedence(tasks[0].var_pointer, duration[0], tasks[1].var_pointer)
    return X



def Mistral_NoOverlap(tasks, durations):
    X = Variable()
    #if getattr(tasks, "__iter__", False):
    X.var_pointer = mistral_lib.CSP._Disjunctive(tasks[0].var_pointer, durations[0][0], tasks[1].var_pointer, durations[0][1], 0)
    #else:
    #    X.var_pointer = mistral_lib.CSP._Disjunctive(tasks.var_pointer, tasks.duration, durations.var_pointer, durations.duration, 0)
    return X

def Mistral_UnaryResource(disjuncts):
    return disjuncts
