/* File : mistral.i */
%module mistral_lib

%{
#include <mistral_sol.h>
using namespace Mistral;
%}


%include "carrays.i"
%array_functions(BuildObject*, VarArray);
%array_functions(int, IntArray);




struct Statistics {
  // STATS
  double            start_time;
  long unsigned int nodes;
  long unsigned int unit_props;
  long unsigned int conflicts;
  unsigned int      literals;
  unsigned int      small;
  double            base_avg_size;
  double            learnt_avg_size;
};

struct Parameters {
  int               seed;
  double            time_limit;
  long unsigned int restart_limit;
  long unsigned int restart_base;
  double            restart_factor;
  double            activity_increment;
  int               policy;
  double            forgetfulness;
  unsigned int      randomization;
  bool              shuffle;
  int               verbosity;
  double            decay;
};

class RestartPolicy;

int atom( const Literal l );

class SatSolver 
{

public:
  /**@name Parameters*/
  //@{
  /// search statistics
  Statistics stats;
  /// search parameters
  Parameters params;
  //@}


  /**@name Constructors*/
  //@{
  SatSolver();
  SatSolver(const char* filename);
  virtual ~SatSolver();

  virtual void init(const int n, const int m);
  void parseDimacs(const char* filename);
  //@}

  /**@name Miscellanous*/
  //@{
  void printAll(std::ostream& o) const;
  void printWatchers(std::ostream& o, int beg=NOVAL, int end=NOVAL) const;
  void printDecisions(std::ostream& o, bool m=true) const;
  void printClauses(std::ostream& o) const;
  void printClause(std::ostream& o, Clause* cl) const;
  //@}

  /**@name Solving Methods*/
  //@{
  /// Solves the problem and set the status accordingly
  virtual int solve();

  /// Add a clause to the base/learnt
  void addClause( Vector<Literal>& conflict );
  //@}
};


class BuildObject
{
public:
  int getValue() const ;
  int getMin() const;
  int getMax() const;
  virtual std::string toString() const ;
};


class CSP {

public:
  // stats
  unsigned int nList;
  unsigned int eList;
  unsigned int nBit;
  unsigned int eBit;
  unsigned int nRange;
  unsigned int eRange;
  unsigned int nBoolean;
  unsigned int eBoolean;
  unsigned int nConstant;
  unsigned int eConstant;
  unsigned int nValues;
  unsigned int eValues;
  unsigned int nConstraint;
  unsigned int eConstraint;
  unsigned int nNeq;

  /*!@name Constructors*/
  //@{
  CSP() ;
  virtual ~CSP() ; 
  //@}

  /*!@name Accessors*/
  //@{
  int numVars() const ;
  int numEVars() const ;
  /// Add a single variable x 
  void add( BuildObject *x ) ;
  void add( BuildObjectObjective *x ) ;
  void add( BuildObject **x, const int n ) ;

  /// Build the data structure used during search;
  void build( Solver *s ) ;
  void build( SatSolver *s ) ;
  bool preBuild( ) ;
  //@}


  /*!@name Utils*/
  //@{
  static BuildObject* _Variable( const int nv );
  static BuildObject* _Variable( const int lb, const int ub );
  static BuildObject* _Variable( const int *array, const int l );
  static BuildObject* _Variable( const int *array, const int l,
				 const int lb, const int ub );
  static BuildObject* _Variable( const BitSet& d, const int l,
				 const int lb, const int ub );


  static BuildObject** _VarArray( const int n, const int nv );
  static BuildObject** _VarArray( const int n, const int lb, const int ub );
  static BuildObject** _VarArray( const int n, const int *array, const int length );
  static BuildObject** _VarArray( const int n, const int *array, const int length, 
				  const int lb, const int ub );
  
  static BuildObject* _Constraint( const int t, BuildObject **x, const int n, int*p );

  static BuildObject* _Negation( BuildObject *x );
  static BuildObject* _Not( BuildObject *x );
  static BuildObject* _Abs( BuildObject *x );
  static BuildObject* _And( BuildObject *x, BuildObject *y );
  static BuildObject* _Or( BuildObject *x, BuildObject *y );
  static BuildObject* _IfThenElse( BuildObject *x, BuildObject *y, BuildObject *z );
  static BuildObject* _Disjunctive( BuildObject *x, const int dx, BuildObject *y, int const dy, const int t );
  static BuildObject* _Precedence( BuildObject *x, const int d, BuildObject *y );
  static BuildObject* _Precedence( const int d, BuildObject *y );
  static BuildObject* _Precedence( BuildObject *x, const int d );
  static BuildObject* _Add( BuildObject *x, BuildObject *y );
  static BuildObject* _Sub( BuildObject *x, BuildObject *y );
  static BuildObject* _Mul( BuildObject *x, BuildObject *y );
  static BuildObject* _Mul( BuildObject *x, const int v );
  static BuildObject* _Div( BuildObject *x, BuildObject *y );
  static BuildObject* _Mod( BuildObject *x, BuildObject *y );
  static BuildObject* _Mod( BuildObject *x, const int m );
  static BuildObject* _Equal( BuildObject *x, BuildObject *y, const int spin );
  static BuildObject* _Equal( BuildObject *x, const int v, const int spin );
  static BuildObject* _Member( BuildObject *x, const BitSet& k, const int spin );
  static BuildObject* _Sum( BuildObject **x, const int nx, const int *os=NULL );
  static BuildObject* _Sum( BuildObject *x, const int o );
  static BuildObject* _Sum( BuildObject *x, BuildObject *y, const int t );
  static BuildObject* _Min( BuildObject *x, BuildObject *y );
  static BuildObject* _Max( BuildObject *x, BuildObject *y );
  static BuildObject* _Min( BuildObject **x, const int n );
  static BuildObject* _Max( BuildObject **x, const int n );
  static BuildObject* _Element( BuildObject **x, const int nx, const int k );
  static BuildObject* _AllDifferent( BuildObject **x, const int nx );
  static BuildObject* _AllDifferent( BuildObject **x, const int nx, const int c, const int k );
  static BuildObject* _Gcc( BuildObject **x, const int nx,
			    const int firstDomainValue, 
			    const int lastDomainValue,
			    const int* minOccurrences,
			    const int* maxOccurrences );
  static BuildObject* _LexOrder( BuildObject **x, const int nx, const int eq=1 );
  static BuildObject* _Cardinality( BuildObject **x, const int nx, const int k );

  static BuildObject* _Table( BuildObject **x, const int nx, const bool sp=true );
  static BuildObject* _Table( BuildObject **x, const int nx, const int tidx, const bool sp );
  static void _addTuple( BuildObject *tab, const int* sol );
    
  static int addTable( const int a, const int n );
  static void addTuple( const int idx, int *tuple );

  static BuildObjectObjective* _Maximise( BuildObject *x );
  static BuildObjectObjective* _Minimise( BuildObject *x );
  //@}
  
  /*!@name Miscellaneous*/
  //@{
  /// Printshort the CSP
  void printPython() const ;
  void printXML(std::ostream&) const ;
  //@}
};



class Solver {

 public:
  int status;

  /// The list of Variables
  Vector < VariableInt* > variables;
  /// The list of Constraints
  Vector < Constraint* > constraints;
  /// Number of nodes, that is recursive calls to  the dfs algo
  unsigned long int NODES; 
  /// Limit on the number of nodes
  unsigned long int NDSLIMIT;
  /// Number of backtracks, that is unsuccesful recursive calls 
  unsigned long int BACKTRACKS;
  /// Limit on the number of backtracks, used for restarting
  unsigned long int BTSLIMIT;
  /// Number of constraint failures
  unsigned long int FAILURES; 
  /// Limit on the number of failures, used for probing
  unsigned long int FAILLIMIT;
  /// Number of calls to a constraint propagator
  unsigned long int PROPAGS;
  /// Number of constraint checks (unused)
  unsigned long int CHECKS;
  /// Number of solutions found so far
  unsigned long int SOLUTIONS;
  /// Number of solutions found so far
  unsigned long int FILTERINGS;
  /// Number of solutions found so far
  unsigned long int RESTORES;
  /// flag to know if there is a limit
  unsigned int LIMIT;
  /// Number of solutions found so far
  unsigned long int MISC;

  /// Number of backtracks for each restart
  Vector<unsigned long int> BTSLIST;

  /// timestamp
  double INITTIME;
  /// timestamp
  double STARTTIME;
  /// timestamp
  double ENDTIME;
  /// timestamp
  double TOTTIME;
  /// timestamp
  double TIMELIMIT;

  /// level of verbosity
  int verbosity;

  /**@name Constructors*/
  //@{
  Solver();
  Solver(CSP& c);
  Solver(CSP& c, BuildObject **x, const int l);
  virtual ~Solver();
  //@}
  

  /**@name Solving methods*/
  //@{
  /*!
    Decomposed dfs
   */
  int startNewSearch();
  /*!
    Decomposed dfs
   */
  int getNextSolution();
  /*!
     Preprocessing step (closure computed with filtering())
  */
  bool preprocess();
   /// Returns UNSAT/SAT/LIMITOUT
  /*!
    presolving method
   */
  int presolve();
  /*!
    Simple solving method
   */
  int solve();
  /*!
    LDS solving method
   */
  int ldSolve( const int step=1, const int limit=NOVAL );
  int ldSolve( const int*, const int step=1, const int limit=NOVAL );
  int ldSolve( const int, BuildObject**, int*, const int step=1, const int limit=NOVAL );
  int ldSolve( ConstraintLDS*, const int step=1, const int limit=NOVAL );
  /// Returns UNSAT/SAT/LIMITOUT
  /*!
    Solve using restarts. There is no randomization besides
    that induced by the variables and values ordering.
   */
  int solve_and_restart( const int policy = GEOMETRIC, 
			 const unsigned int base = 32, 
			 const double factor = 1.3333333,
			 const double decay = 0.0 ); 
  /*!
    Diarmuid's random probes
   */
  int random_probe( const unsigned int iterations = 100, 
		    const unsigned int limit = 30 );
  /*!
    singleton arc consistency
   */
  int sacPreprocess( const bool complete = false, 
		     const int wtype = (Weighter::IPT | Weighter::WLD) );
  //@}
  

  /**@name Accessors*/
  //@{
    //void setHeuristic(string hname, const int rdz);
  void setHeuristic(const char* hname, const char* vname, const int rdz);	
  void setBranching(BuildObject **x, const int l, const char* val_name);
   /// Cutoff in number of backtracks     
  void setNodeLimit     (unsigned long nl){ NDSLIMIT  = nl; LIMIT = 1; }
  /// Cutoff in number of backtracks     
  void setBacktrackLimit(unsigned long nl){ BTSLIMIT  = nl; LIMIT = 1; }
  /// Cutoff in number of failures  
  void setFailureLimit  (unsigned long nl){ FAILLIMIT = nl; LIMIT = 1; }
  /// Cutoff in cpu time (s)
  void setTimeLimit     (double tl)       { TIMELIMIT = tl; LIMIT = 1; }  
  /// Print satisfiability, number of backtracks and cpu time
  void setVerbosity     (const int v)     { verbosity =  v; }  
  /// Add a learner
  Weighter* setLearner( int wtype );
  /// Use domain splitting
  void setDomainSplitting() { domainSplitting = true; }
  /// Shuffle the array of variables
  void randomizeSequence();
  /// Randomize the array of variables before restarts
  void setRandomized( const bool r=true ) { randomizedRestart = r; }
  /// Set the random seed
  void setRandomSeed( const unsigned int seed ) { usrand(seed); }
  /// Setup the nogoods on restart
  void setRestartNogood();
  /// Setup the forget utility for the clause base
  void setForgetfulness( const double f );
  /// Print statistics corresponding to the bitset stats
  void printStatistics(std::ostream &log=std::cout, const int stats=ALL);
  unsigned long int getBacktracks() const ;
  unsigned long int getNodes     () const ;
  unsigned long int getFailures  () const ;
  unsigned long int getChecks    () const ;
  unsigned long int getPropags   () const ;
  double            getTime      () const ;
  void printPython() const;
  void printXML(std::ostream& o) const;
  void printWeightProfile(std::ostream& o, int limit, int t) const;
  //@}

};

